@@ -11,6 +11,7 @@
 #include "clock-face.h"
 #include "clock-location-tile.h"
 #include "clock-location.h"
+#include "clock-utils.h"
 #include "clock-marshallers.h"
 #include "set-timezone.h"
 
@@ -18,9 +19,7 @@ G_DEFINE_TYPE (ClockLocationTile, clock_location_tile, GTK_TYPE_ALIGNMENT)
 
 enum {
 	TILE_PRESSED,
-	TIMEZONE_SET,
-	WEATHER_UPDATED,
-	NEED_FORMATTED_TIME,
+	NEED_CLOCK_FORMAT,
 	LAST_SIGNAL
 };
 
@@ -42,6 +41,8 @@ typedef struct {
         GtkWidget *current_button;
         GtkWidget *current_label;
         GtkWidget *current_marker;
+        GtkWidget *current_spacer;
+        GtkSizeGroup *current_group;
         GtkSizeGroup *button_group;
 
         GtkWidget *weather_icon;
@@ -104,32 +105,14 @@ clock_location_tile_class_init (ClockLocationTileClass *this_class)
 					      NULL,
 					      g_cclosure_marshal_VOID__VOID,
 					      G_TYPE_NONE, 0);
-	signals[TILE_PRESSED] = g_signal_new ("timezone-set",
-					      G_TYPE_FROM_CLASS (g_obj_class),
-					      G_SIGNAL_RUN_FIRST,
-					      G_STRUCT_OFFSET (ClockLocationTileClass, timezone_set),
-					      NULL,
-					      NULL,
-					      g_cclosure_marshal_VOID__VOID,
-					      G_TYPE_NONE, 0);
-	signals[WEATHER_UPDATED] = g_signal_new ("weather-updated",
-						 G_TYPE_FROM_CLASS (g_obj_class),
-						 G_SIGNAL_RUN_FIRST,
-						 G_STRUCT_OFFSET (ClockLocationTileClass, weather_updated),
-						 NULL,
-						 NULL,
-						 _clock_marshal_VOID__OBJECT_STRING,
-						 G_TYPE_NONE, 2,
-						 G_TYPE_OBJECT,
-						 G_TYPE_STRING);
-	signals[NEED_FORMATTED_TIME] = g_signal_new ("need-formatted-time",
-						     G_TYPE_FROM_CLASS (g_obj_class),
-						     G_SIGNAL_RUN_LAST,
-						     G_STRUCT_OFFSET (ClockLocationTileClass, need_formatted_time),
-						     NULL,
-						     NULL,
-						     _clock_marshal_STRING__VOID,
-						     G_TYPE_STRING, 0);
+	signals[NEED_CLOCK_FORMAT] = g_signal_new ("need-clock-format",
+						   G_TYPE_FROM_CLASS (g_obj_class),
+						   G_SIGNAL_RUN_LAST,
+						   G_STRUCT_OFFSET (ClockLocationTileClass, need_clock_format),
+						   NULL,
+						   NULL,
+						   _clock_marshal_INT__VOID,
+						   G_TYPE_INT, 0);
 }
 
 static void
@@ -167,6 +150,11 @@ clock_location_tile_finalize (GObject *g_obj)
                 priv->button_group = NULL;
         }
 
+        if (priv->current_group) {
+                g_object_unref (priv->current_group);
+                priv->current_group = NULL;
+        }
+
         G_OBJECT_CLASS (clock_location_tile_parent_class)->finalize (g_obj);
 }
 
@@ -181,16 +169,11 @@ press_on_tile      (GtkWidget             *widget,
 }
 
 static void
-make_current (GtkWidget *widget, ClockLocationTile *tile)
+make_current_cb (gpointer data, GError *error)
 {
-        ClockLocationTilePrivate *priv = PRIVATE (tile);
-        GError *error = NULL;
-        GtkWidget *dialog;
+	GtkWidget *dialog;
 
-        if (clock_location_make_current (priv->location, &error)) {
-		g_signal_emit (tile, signals[TIMEZONE_SET], 0);
-        }
-        else if (error) {
+        if (error) {
                 dialog = gtk_message_dialog_new (NULL,
                                                  0,
                                                  GTK_MESSAGE_ERROR,
@@ -205,32 +188,65 @@ make_current (GtkWidget *widget, ClockLocationTile *tile)
         }
 }
 
+static void
+make_current (GtkWidget *widget, ClockLocationTile *tile)
+{
+        ClockLocationTilePrivate *priv = PRIVATE (tile);
+
+	clock_location_make_current (priv->location, 
+				     (GFunc)make_current_cb, tile, NULL);
+}
+
 static gboolean
 enter_or_leave_tile (GtkWidget             *widget,
                      GdkEventCrossing      *event,
                      ClockLocationTile *tile)
 {
-        ClockLocationTilePrivate *priv = PRIVATE (tile);
+	ClockLocationTilePrivate *priv = PRIVATE (tile);
 
-        if (event->type == GDK_ENTER_NOTIFY) {
-	      	gint can_set;
+	if (event->mode != GDK_CROSSING_NORMAL) {
+		return TRUE;
+	}
 
-		can_set = can_set_system_timezone ();
-		if (!clock_location_is_current (priv->location) &&
-		    can_set != 0) {
+	if (clock_location_is_current (priv->location)) {
+		gtk_widget_hide (priv->current_button);
+		gtk_widget_hide (priv->current_spacer);
+		gtk_widget_show (priv->current_marker);
+
+		return TRUE;
+	}
+
+	if (event->type == GDK_ENTER_NOTIFY) {
+		gint can_set;
+
+		if (clock_location_is_current_timezone (priv->location))
+			can_set = 2;
+		else
+			can_set = can_set_system_timezone ();
+		if (can_set != 0) {
 			gtk_label_set_markup (GTK_LABEL (priv->current_label),
-					    can_set == 1 ?
-					    	_("<small>Set...</small>") :
-					    	_("<small>Set</small>"));
+						can_set == 1 ?
+							_("<small>Set...</small>") :
+							_("<small>Set</small>"));
+			gtk_widget_hide (priv->current_spacer);
+			gtk_widget_hide (priv->current_marker);
 			gtk_widget_show (priv->current_button);
 		}
-       }
-       else {
-               if (event->detail != GDK_NOTIFY_INFERIOR)
-                       gtk_widget_hide (priv->current_button);
-       }
+		else {
+			gtk_widget_hide (priv->current_marker);
+			gtk_widget_hide (priv->current_button);
+			gtk_widget_show (priv->current_spacer);
+		}
+	}
+	else {
+		if (event->detail != GDK_NOTIFY_INFERIOR) {
+			gtk_widget_hide (priv->current_button);
+			gtk_widget_hide (priv->current_marker);
+			gtk_widget_show (priv->current_spacer);
+		}
+	}
 
-       return TRUE;
+	return TRUE;
 }
 
 static void
@@ -240,6 +256,10 @@ clock_location_tile_fill (ClockLocationTile *this)
 	GtkWidget *align;
         GtkWidget *strut;
         GtkWidget *box;
+        GtkWidget *alignment;
+        GtkWidget *tile;
+        GtkWidget *head_section;
+	gint can_set;
 
         priv->box = gtk_event_box_new ();
 
@@ -251,11 +271,11 @@ clock_location_tile_fill (ClockLocationTile *this)
         g_signal_connect (priv->box, "leave-notify-event",
                           G_CALLBACK (enter_or_leave_tile), this);
 
-        GtkWidget *alignment = gtk_alignment_new (0, 0, 1, 0);
+        alignment = gtk_alignment_new (0, 0, 1, 0);
         gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 3, 3, 3, 0);
 
-        GtkWidget *tile = gtk_hbox_new (FALSE, 6);
-        GtkWidget *head_section = gtk_vbox_new (FALSE, 0);
+        tile = gtk_hbox_new (FALSE, 6);
+        head_section = gtk_vbox_new (FALSE, 0);
 
         priv->city_label = gtk_label_new (NULL);
         gtk_misc_set_alignment (GTK_MISC (priv->city_label), 0, 0);
@@ -277,26 +297,58 @@ clock_location_tile_fill (ClockLocationTile *this)
         gtk_box_pack_start (GTK_BOX (box), align, FALSE, FALSE, 0);
         gtk_box_pack_start (GTK_BOX (box), priv->time_label, FALSE, FALSE, 0);
 
+	can_set = can_set_system_timezone ();
+
         priv->current_button = gtk_button_new ();
 	priv->current_label = gtk_label_new ("");
+	if (can_set != 0) {
+		gtk_label_set_markup (GTK_LABEL (priv->current_label),
+				can_set == 1 ?  
+					_("<small>Set...</small>") :
+					_("<small>Set</small>"));
+	}
         gtk_widget_show (priv->current_label);
         gtk_widget_set_no_show_all (priv->current_button, TRUE);
         gtk_container_add (GTK_CONTAINER (priv->current_button), priv->current_label);
+	/* FIXME: this needs to talk about locations, not timezones */
         gtk_widget_set_tooltip_text (priv->current_button, _("Set as current timezone for this computer"));
 
 	priv->current_marker = gtk_image_new_from_icon_name ("go-home", GTK_ICON_SIZE_BUTTON);
 	gtk_misc_set_alignment (GTK_MISC (priv->current_marker), 1.0, 0.5);
 	gtk_widget_set_no_show_all (priv->current_marker, TRUE);
 
+	priv->current_spacer = gtk_event_box_new ();
+	gtk_widget_set_no_show_all (priv->current_spacer, TRUE);
+
         strut = gtk_event_box_new ();
         gtk_box_pack_start (GTK_BOX (box), strut, TRUE, TRUE, 0);
         gtk_box_pack_start (GTK_BOX (box), priv->current_button, FALSE, FALSE, 0);
         gtk_box_pack_start (GTK_BOX (box), priv->current_marker, FALSE, FALSE, 0);
+        gtk_box_pack_start (GTK_BOX (box), priv->current_spacer, FALSE, FALSE, 0);
         priv->button_group = gtk_size_group_new (GTK_SIZE_GROUP_VERTICAL);
         gtk_size_group_set_ignore_hidden (priv->button_group, FALSE);
         gtk_size_group_add_widget (priv->button_group, strut);
         gtk_size_group_add_widget (priv->button_group, priv->current_button);
 
+	/* 
+	 * Avoid resizing the popup as the tiles display the current marker, 
+	 * set button or nothing. For that purpose, replace 'nothing' with 
+	 * an event box, and force the button, marker and spacer to have the 
+	 * same size via a size group. The visibility of the three is managed
+ 	 * manually to ensure that only one of them is shown at any time. 
+ 	 * (The all have to be shown initially to get the sizes worked out, 
+ 	 * but they are never visible together). 
+	 */
+        priv->current_group = gtk_size_group_new (GTK_SIZE_GROUP_BOTH);
+        gtk_size_group_set_ignore_hidden (priv->current_group, FALSE);
+        gtk_size_group_add_widget (priv->current_group, priv->current_button);
+        gtk_size_group_add_widget (priv->current_group, priv->current_marker);
+        gtk_size_group_add_widget (priv->current_group, priv->current_spacer);
+	
+	gtk_widget_show (priv->current_button);
+	gtk_widget_show (priv->current_marker);
+	gtk_widget_show (priv->current_spacer);
+
         g_signal_connect (priv->current_button, "clicked",
                           G_CALLBACK (make_current), this);
 
@@ -310,7 +362,7 @@ clock_location_tile_fill (ClockLocationTile *this)
         gtk_container_add (GTK_CONTAINER (priv->box), alignment);
         gtk_container_add (GTK_CONTAINER (this), priv->box);
 
-        clock_location_tile_refresh (this);
+        clock_location_tile_refresh (this, TRUE);
 }
 
 static gboolean
@@ -372,46 +424,115 @@ copy_tm (struct tm *from, struct tm *to)
         to->tm_yday = from->tm_yday;
 }
 
-static void
-emit_weather_updated (ClockLocationTile *this, GdkPixbuf *weather_icon, const char *temperature)
+static char *
+format_time (struct tm   *now, 
+             char        *tzname,
+             ClockFormat  clock_format,
+	     long         offset)
 {
-	g_signal_emit (this, signals[WEATHER_UPDATED], 0, weather_icon, temperature);
+	char buf[256];
+	char *format;
+	time_t local_t;
+	struct tm local_now;
+	char *utf8;	
+	char *tmp;	
+	long hours, minutes;
+
+	time (&local_t);
+	localtime_r (&local_t, &local_now);
+
+	if (local_now.tm_wday != now->tm_wday) {
+		if (clock_format == CLOCK_FORMAT_12) {
+/* translators: This is a strftime format string that is used for formatting 
+ * the time that is displayed for locations in the popup. This string is used
+ * for 12-hour format (the %p expands to am/pm), when the local weekday differs
+ * from the weekday at the location (the %A expands to the weekday).
+ * There should be little need to translate this string.
+ */
+			format = _("%l:%M <small>%p (%A)</small>");
+		}
+		else {
+/* translators: This is a strftime format string that is used for formatting 
+ * the time that is displayed for locations in the popup. This string is used
+ * for 24-hour format (the %p expands to am/pm), when the local weekday differs
+ * from the weekday at the location (the %A expands to the weekday).
+ * There should be little need to translate this string.
+ */
+			format = _("%H:%M <small>(%A)</small>");
+		}
+	}
+	else {
+		if (clock_format == CLOCK_FORMAT_12) {
+/* translators: This is a strftime format string that is used for formatting 
+ * the time that is displayed for locations in the popup. This string is used
+ * for 12-hour format (the %p expands to am/pm). 
+ * There should be little need to translate this string.
+ */
+			format = _("%l:%M <small>%p</small>");
+		}
+		else {
+/* translators: This is a strftime format string that is used for formatting 
+ * the time that is displayed for locations in the popup. This string is used
+ * for 24-hour format.
+ * There should be little need to translate this string.
+ */
+			format = _("%H:%M");
+		}
+	}
+
+	if (strftime (buf, sizeof (buf), format, now) <= 0) {
+		strcpy (buf, "???");
+	}
+
+        hours = offset / 3600;
+        minutes = labs (offset % 3600) / 60;
+
+	if (hours != 0 && minutes != 0) {
+		tmp = g_strdup_printf ("%s <small>%s %+ld:%ld</small>", buf, tzname, hours, minutes);
+	}
+	else if (hours != 0) {
+		tmp = g_strdup_printf ("%s <small>%s %+ld</small>", buf, tzname, hours);
+	}
+	else {
+		tmp = g_strdup_printf ("%s <small>%s</small>", buf, tzname);
+	}
+
+	utf8 = g_locale_to_utf8 (tmp, -1, NULL, NULL, NULL);
+
+	g_free (tmp);
+
+	return utf8;
 }
 
 void
-clock_location_tile_refresh (ClockLocationTile *this)
+clock_location_tile_refresh (ClockLocationTile *this, gboolean force_refresh)
 {
         ClockLocationTilePrivate *priv = PRIVATE (this);
-        gchar *tmp, *tmp2, *tzname;
+        gchar *tmp, *tzname;
         struct tm now;
-	long offset, hours, minutes;
+	long offset;
+	int format;
 
 	g_return_if_fail (IS_CLOCK_LOCATION_TILE (this));
 
         if (clock_location_is_current (priv->location)) {
-		if (!GTK_WIDGET_VISIBLE (priv->current_marker)) {
-			GdkPixbuf *pixbuf;
-                        const gchar *temp = NULL;
-
-			pixbuf = gtk_image_get_pixbuf (GTK_IMAGE (priv->weather_icon));
-                        if (clock_location_get_weather_info (priv->location))
-                               temp = weather_info_get_temp_summary (clock_location_get_weather_info (priv->location));
-
-			emit_weather_updated (this, pixbuf, temp);
+		gtk_widget_hide (priv->current_spacer);
+		gtk_widget_hide (priv->current_button);
+		gtk_widget_show (priv->current_marker);
+	}
+	else {
+		if (GTK_WIDGET_VISIBLE (priv->current_marker)) {
+			gtk_widget_hide (priv->current_marker);
+			gtk_widget_hide (priv->current_button);
+			gtk_widget_show (priv->current_spacer);
 		}
-
-                gtk_widget_hide (priv->current_button);
-                gtk_widget_show (priv->current_marker);
-        }
-        else {
-                gtk_widget_hide (priv->current_marker);
 	}
 
         if (clock_needs_face_refresh (this)) {
                 clock_face_refresh (CLOCK_FACE (priv->clock_face));
         }
 
-        if (!clock_needs_label_refresh (this)) {
+        if (!force_refresh && !clock_needs_label_refresh (this)) {
                 return;
         }
 
@@ -426,36 +547,27 @@ clock_location_tile_refresh (ClockLocationTile *this)
         gtk_label_set_markup (GTK_LABEL (priv->city_label), tmp);
         g_free (tmp);
 
-	tmp = NULL;
-	g_signal_emit (this, signals[NEED_FORMATTED_TIME], 0, &tmp);
+	g_signal_emit (this, signals[NEED_CLOCK_FORMAT], 0, &format);
 
 	offset = - priv->last_offset;
 
-	hours = offset / 3600;
-	minutes = labs (offset % 3600) / 60;
+	tmp = format_time (&now, tzname, format, offset);
 
-	if (hours != 0 && minutes != 0)
-		tmp2 = g_strdup_printf ("%s  <small>%+ld:%ld</small>", tmp ? tmp : "", hours, minutes);
-	else if (hours != 0)
-		tmp2 = g_strdup_printf ("%s  <small>%+ld</small>", tmp ? tmp : "", hours);
-	else {
-		tmp2 = tmp;
-		tmp = NULL;
-	}
+        gtk_label_set_markup (GTK_LABEL (priv->time_label), tmp);
 
-        gtk_label_set_markup (GTK_LABEL (priv->time_label), tmp2);
         g_free (tmp);
-	g_free (tmp2);
 }
 
 void
-weather_info_setup_tooltip (WeatherInfo *info, GtkTooltip *tooltip)
+weather_info_setup_tooltip (WeatherInfo *info, ClockLocation *location, GtkTooltip *tooltip)
 {
         GdkPixbuf *pixbuf = NULL;
         GtkIconTheme *theme = NULL;
-	const gchar *conditions, *temp, *apparent, *wind;
+	const gchar *conditions, *wind;
+	gchar *temp, *apparent;
 	gchar *line1, *line2, *line3, *line4, *tip;
 	const gchar *icon_name;
+	const gchar *sys_timezone;
 
        	icon_name = weather_info_get_icon_name (info);
         theme = gtk_icon_theme_get_default ();
@@ -471,14 +583,18 @@ weather_info_setup_tooltip (WeatherInfo *info, GtkTooltip *tooltip)
 	else
 		line1 = g_strdup (weather_info_get_sky (info));
 
-	temp = weather_info_get_temp_summary (info);
-	apparent = weather_info_get_apparent (info);
+	/* we need to g_strdup() since both functions return the same address
+	 * of a static buffer */
+	temp = g_strdup (weather_info_get_temp (info));
+	apparent = g_strdup (weather_info_get_apparent (info));
 	if (strcmp (apparent, temp) != 0 &&
 	    /* FMQ: it's broken to read from another module's translations; add some API to libgweather. */
             strcmp (apparent, dgettext ("gnome-applets-2.0", "Unknown")) != 0)
 		line2 = g_strdup_printf (_("%s, feels like %s"), temp, apparent);
 	else
 		line2 = g_strdup (temp);
+	g_free (temp);
+	g_free (apparent);
 
 	wind = weather_info_get_wind (info);
         if (strcmp (apparent, dgettext ("gnome-applets-2.0", "Unknown")) != 0)
@@ -486,9 +602,20 @@ weather_info_setup_tooltip (WeatherInfo *info, GtkTooltip *tooltip)
 	else
 		line3 = g_strdup ("");
 
+	/* FIXME: we need libgweather to give us the time,
+	 * not just a formatted string
+	 */
+	sys_timezone = getenv ("TZ");
+	setenv ("TZ", clock_location_get_timezone (location), 1);
+	tzset ();
 	line4 = g_strdup_printf (_("Sunrise: %s / Sunset: %s"),
 				 weather_info_get_sunrise (info),
 				 weather_info_get_sunset (info));
+	if (sys_timezone)
+		setenv ("TZ", sys_timezone, 1);
+	else
+		unsetenv ("TZ");
+	tzset ();
 
 	tip = g_strdup_printf ("<b>%s</b>\n%s\n%s%s", line1, line2, line3, line4);
 	gtk_tooltip_set_markup (tooltip, tip);
@@ -516,7 +643,7 @@ weather_tooltip (GtkWidget  *widget,
 	if (!info || !weather_info_is_valid (info))
 		return FALSE;
 
-	weather_info_setup_tooltip (info, tooltip);
+	weather_info_setup_tooltip (info, priv->location, tooltip);
 
 	return TRUE;
 }
@@ -529,7 +656,6 @@ update_weather_icon (ClockLocation *loc, WeatherInfo *info, gpointer data)
         GdkPixbuf *pixbuf = NULL;
         GtkIconTheme *theme = NULL;
         const gchar *icon_name;
-        const gchar *temp = NULL;
 
         if (!info || !weather_info_is_valid (info))
                 return;
@@ -538,14 +664,9 @@ update_weather_icon (ClockLocation *loc, WeatherInfo *info, gpointer data)
         theme = gtk_icon_theme_get_default ();
         pixbuf = gtk_icon_theme_load_icon (theme, icon_name, 16, 0, NULL);
 
-        temp = weather_info_get_temp_summary (info);
-
         if (pixbuf) {
                 gtk_image_set_from_pixbuf (GTK_IMAGE (priv->weather_icon), pixbuf);
                 gtk_alignment_set_padding (GTK_ALIGNMENT (gtk_widget_get_parent (priv->weather_icon)), 0, 0, 0, 6);
-		if (clock_location_is_current (loc)) {
-			emit_weather_updated (tile, pixbuf, temp);
-		}
         }
 }
 
